from dataclasses import dataclass
from snowday.exceptions import InvalidWhitespaceException
from snowday.util import get_resource_type, settings_str_from


@dataclass
class BaseSnowflakeObject:

    name: str

    def __post_init__(self):
        if " " in self.name:
            raise InvalidWhitespaceException(
                f"Invalid whitespace in name: {self.name}\n\nPlease make it {self.name.replace(' ', '')} instead."
            )
        # Ensure names are uppercased.
        self.name = self.name.upper()
        if callable(getattr(self, "_validate", None)):
            self._validate()

    @property
    def _settings_str(self):
        return settings_str_from(self)

    @property
    def fqn(self):
        return self.name

    @property
    def resource_type(self):
        return get_resource_type(self)

    @property
    def describe(self):
        return f"describe {self.resource_type} {self.fqn};"

    @property
    def drop(self):
        return f"drop {self.resource_type} {self.fqn};"

    @property
    def drop_if_exists(self):
        return f"drop {self.resource_type} if exists {self.fqn};"

    @property
    def create(self):
        if hasattr(self, "__base__"):  # FIXME: figure out a way to do this better
            raise NotImplementedError("create is not implemented on base classes")
        return f"create {self.resource_type} " f"{self.fqn} " f"{self._settings_str};"

    @property
    def create_if_not_exists(self):
        return (
            f"create {self.resource_type} if not exists "
            f"{self.fqn} "
            f"{self._settings_str};"
        )

    @property
    def create_or_replace(self):
        return (
            f"create or replace {self.resource_type} "
            f"{self.fqn} "
            f"{self._settings_str};"
        )



from dataclasses import dataclass, field
from typing import List

from snowday.objects.base import BaseSnowflakeObject
from snowday.exceptions import (
    InvalidParamsException,
    InvalidWhitespaceException,
    UnsupportedFeatureException,
)


# https://docs.snowflake.com/en/sql-reference/sql/create.html


@dataclass
class Account:

    name: str
    admin_name: str
    admin_password: str
    email: str
    edition: str  # either STANDARD or ENTERPRISE or BUSINESS_CRITICAL
    region: str
    comment: str

    def __post_init__(self):
        # Make sure edition is valid
        if self.edition not in ["STANDARD", "ENTERPRISE", "BUSINESS_CRITICAL"]:
            raise InvalidParamsException(
                "Edition must be STANDARD, ENTERPRISE, or BUSINESS_CRITICAL"
            )
        # Make sure region is valid. FIXME!
        # Ensure names are uppercased.
        self.name = self.name.upper()

    @property
    def create(self):
        return (
            f"create account {self.name} "
            f"admin_name={self.admin_name}, admin_password='{self.admin_password}', "
            f"email='{self.email}', edition={self.edition}, "
            f"region='{self.region}' comment='{self.comment}';"
        )

    @property
    def drop(self):
        return f"drop account {self.name};"

    @property
    def enable_replication(self):
        return f"select system$global_account_set_parameter('{self.name}', 'ENABLE_ACCOUNT_DATABASE_REPLICATION', 'true');"

    @property
    def disable_replication(self):
        return f"select system$global_account_set_parameter('{self.name}', 'ENABLE_ACCOUNT_DATABASE_REPLICATION', 'false');"


@dataclass
class AccountObject(BaseSnowflakeObject):
    pass


@dataclass
class ManagedAccount(BaseSnowflakeObject):
    admin_name: str
    admin_password: str
    comment: str = ""
    type: str = "READER"

    @property
    def resource_type(self):
        return "managed account"

    @property
    def describe(self):
        raise NotImplementedError("describe is not supported on managed accounts")

    @property
    def create_if_not_exists(self):
        raise NotImplementedError(
            "create if not exists is not supported on managed accounts"
        )

    @property
    def create_or_replace(self):
        raise NotImplementedError(
            "create or replace is not supported on managed accounts"
        )


@dataclass
class AwsApiIntegration(AccountObject):
    api_aws_role_arn: str = ""  # Figure out a cleaner way to do this
    api_provider: str = "aws_api_gateway"
    api_allowed_prefixes: List[str] = field(default_factory=list)
    api_blocked_prefixes: List[str] = field(default_factory=list)
    comment: str = ""
    enabled: bool = True  # enabled by default

    @property
    def resource_type(self):
        return "api integration"


@dataclass
class NetworkPolicy(AccountObject):
    allowed_ips: List[str] = field(default_factory=list)
    blocked_ips: List[str] = field(default_factory=list)
    comment: str = ""

    @property
    def create_if_not_exists(self):
        raise UnsupportedFeatureException(
            "Create if not exists for snowflake network policies is unsupported."
        )


@dataclass
class ResourceMonitorTrigger:
    on_percent: int
    do: str  # SUSPEND | SUSPEND_IMMEDIATE | NOTIFY

    @property
    def sql(self):
        return f"on {self.on_percent} percent do {self.do}"


@dataclass
class ResourceMonitor(AccountObject):
    triggers: List[ResourceMonitorTrigger]
    credit_quota: int = 1
    frequency: str = "never"  # must be MONTHLY | DAILY | WEEKLY | YEARLY | NEVER
    start_timestamp: str = "immediately"  # Use strings because.... timestamp | "IMMEDIATELY"
    end_timestamp: str = None  # String whyyy

    @property
    def _settings_str(self):
        return (
            f"with credit_quota={self.credit_quota} "
            f"frequency={self.frequency} "
            f"start_timestamp={self.start_timestamp} "
            f"triggers {self._format_triggers()}"
        )

    def _format_triggers(self):
        return " ".join([trigger.sql for trigger in self.triggers])

    @property
    def create_if_not_exists(self):
        raise UnsupportedFeatureException(
            "Create if not exists for snowflake resource monitors is unsupported."
        )


@dataclass
class Share(AccountObject):
    comment: str = ""

    @property
    def create_if_not_exists(self):
        raise UnsupportedFeatureException(
            "Create if not exists for snowflake shares is unsupported."
        )


@dataclass
class Warehouse(AccountObject):
    warehouse_size: str = "xsmall"  # enum this
    min_cluster_count: int = 1
    max_cluster_count: int = 1
    scaling_policy: str = None  # enum this
    auto_suspend: int = 60
    auto_resume: bool = True
    initially_suspended: bool = True
    max_concurrency_level: int = 10
    statement_queued_timeout_in_seconds: int = 600
    statement_timeout_in_seconds: int = 6000
    resource_monitor: ResourceMonitor = None
    comment: str = ""

    def __validate(self):
        # Validate warehouse size param
        if self.warehouse_size.lower() not in [
            "xsmall",
            "small",
            "medium",
            "large",
            "xlarge",
            "xxlarge",
            "xxxlarge",
        ]:
            raise InvalidParamsException(
                f"{self.warehouse_size} is not a valid snowflake warehouse size"
            )
        # Validate scaling policy param
        # Validate cluster count (if not enterprise account) param


@dataclass
class Role(AccountObject):
    comment: str = ""


@dataclass
class User(BaseSnowflakeObject):
    email: str = None
    password: str = None
    login_name: str = None
    display_name: str = None
    first_name: str = None
    middle_name: str = None
    last_name: str = None
    must_change_password: bool = True
    disabled: bool = False
    snowflake_support: bool = False
    days_to_expiry: int = None
    mins_to_unlock: int = None
    default_warehouse: Warehouse = None
    default_role: Role = None
    default_namespace: str = None
    # ext_authn_duo: bool = False # NOTE! CANNOT BE SET DIRECTLY
    # ext_authn_uid: str = None # # NOTE! CANNOT BE SET DIRECTLY
    min_to_bypass_mfa: int = None
    disable_mfa: bool = None
    rsa_public_key: str = None
    rsa_public_key_2: str = None
    abort_detached_query: bool = True
    autocommit: bool = False
    date_input_format: str = None
    date_output_format: str = None
    error_on_nondeterministic_merge: bool = None
    error_on_nondeterministic_update: bool = None
    lock_timeout: int = None
    query_tag: str = None
    rows_per_resultset: int = None
    statement_timeout_in_seconds: int = None
    timestamp_day_is_always_24h: bool = None
    timestamp_input_format: str = None
    timestamp_ltz_output_format: str = None
    timestamp_ntz_output_format: str = None
    timestamp_output_format: str = None
    timestamp_type_mapping: str = None
    timestamp_tz_output_format: str = None
    timezone: str = None  # restrict to list of valid timezones
    time_input_format: str = None
    time_output_format: str = None
    transaction_default_isolation_level: str = None
    two_digit_century_start: int = None
    unsupported_ddl_action: str = None
    use_cached_result: bool = None
    comment: str = ""


@dataclass
class AzureNotificationIntegration(AccountObject):
    azure_storage_queue_primary_uri: str
    azure_tenant_id: str
    enabled: bool = True
    type: str = "queue"
    notification_provider: str = "azure_storage_queue"
    comment: str = ""

    @property
    def resource_type(self):
        return "notification integration"


@dataclass
class SecurityIntegrationOAuth(AccountObject):
    external_oauth_type: str
    external_oauth_issuer: str
    external_oauth_token_user_mapping_claim: str
    external_oauth_snowflake_user_mapping_attribute: str  # login_name or email
    enabled: bool = True
    external_oauth_jws_keys_url: str = None  # optional
    external_oauth_rsa_public_key: str = None  # optional pk1
    external_oauth_rsa_public_key_2: str = None  # optional pk2
    external_oauth_audience_list: str = None  # optional
    type: str = "external_oauth"

    def __post_init__(self):
        if self.external_oauth_type not in ["okta", "azure", "ping_federate", "custom"]:
            raise InvalidParamsException(
                f"{self.external_oauth_type} is not a valid oauth type."
            )

    @property
    def resource_type(self):
        return "security integration"


@dataclass
class SecurityIntegrationOAuthPartnerApp(AccountObject):
    def __post_init__(self):
        raise NotImplementedError(
            "Snowflake Integration for OAuth Partner Apps not implemented. If you'd like to contribute this functionality, it would be wholly welcomed."
        )

    @property
    def resource_type(self):
        return "security integration"


@dataclass
class SecurityIntegrationOAuthCustom(AccountObject):
    def __post_init__(self):
        raise NotImplementedError(
            "Snowflake Integration for custom OAuth not implemented. If you'd like to contribute this functionality, it would be wholly welcomed."
        )

    @property
    def resource_type(self):
        return "security integration"


@dataclass
class SecurityIntegrationSAML2(AccountObject):
    saml2_issuer: str
    saml2_sso_url: str
    saml2_provider: str
    saml2_x509_cert: str
    enabled: bool = True
    saml2_sp_initiated_login_page_label: str = None
    saml2_enable_sp_initiated: str = None
    saml2_snowflake_x509_cert: str = None
    saml2_sign_request: str = None
    saml2_requested_nameid_format: str = None
    type: str = "saml2"

    @property
    def resource_type(self):
        return "security integration"


@dataclass
class SecurityIntegrationSCIM(AccountObject):
    scim_client: str  # okta, azure, or custom
    run_as_role: str  # okta_provisioner, aad_provisioner, generic_scim_provisioner
    network_policy: str = None
    type: str = "saml2"

    @property
    def resource_type(self):
        return "security integration"


@dataclass
class BaseStorageIntegration(AccountObject):

    storage_allowed_locations: List[str]  # required
    storage_blocked_locations: List[str] = field(default_factory=list)
    comment: str = ""
    enabled: bool = True
    type: str = "external_stage"

    @property
    def resource_type(self):
        return "storage integration"


@dataclass
class StorageIntegrationS3(BaseStorageIntegration):
    storage_aws_role_arn: str = ""  # defaulting this is a dirty hack to get around TypeError: non-default argument 'storage_aws_role_arn' follows default argument
    storage_provider: str = "s3"


@dataclass
class StorageIntegrationGCS(BaseStorageIntegration):
    storage_provider: str = "gcs"


@dataclass
class StorageIntegrationAzure(BaseStorageIntegration):
    azure_tenant_id: str = ""
    storage_provider: str = "azure"



import atexit
from contextlib import contextmanager
from typing import Union, Iterator

import snowflake.connector as sf_connector

from snowday.logger import get_logger
from snowday.objects.account import (
    Account,
    ManagedAccount,
    AwsApiIntegration,
    NetworkPolicy,
    ResourceMonitor,
    Share,
    Warehouse,
    Role,
    User,
    AzureNotificationIntegration,
    SecurityIntegrationOAuth,
    SecurityIntegrationOAuthPartnerApp,
    SecurityIntegrationOAuthCustom,
    SecurityIntegrationSAML2,
    SecurityIntegrationSCIM,
    StorageIntegrationS3,
    StorageIntegrationGCS,
    StorageIntegrationAzure,
)
from snowday.objects.database import (
    Database,
    DatabaseFromShare,
    DatabaseReplica,
    Schema,
)
from snowday.objects.schema import (
    Integration,
    Stage,
    Task,
    View,
    MaterializedView,
    MaskingPolicy,
    FileFormat,
    InternalStage,
    ExternalStage,
    Pipe,
    Stream,
    Function,
    Procedure,
    Sequence,
)
from snowday.objects.table import Table

LOG = get_logger()

ALL_OBJS = Union[
    Account,
    ManagedAccount,
    AwsApiIntegration,
    NetworkPolicy,
    ResourceMonitor,
    Share,
    Warehouse,
    Role,
    User,
    AzureNotificationIntegration,
    SecurityIntegrationOAuth,
    SecurityIntegrationOAuthPartnerApp,
    SecurityIntegrationOAuthCustom,
    SecurityIntegrationSAML2,
    SecurityIntegrationSCIM,
    StorageIntegrationS3,
    StorageIntegrationGCS,
    StorageIntegrationAzure,
    Database,
    DatabaseFromShare,
    DatabaseReplica,
    Schema,
    Integration,
    Stage,
    Task,
    View,
    MaterializedView,
    MaskingPolicy,
    FileFormat,
    InternalStage,
    ExternalStage,
    Pipe,
    Stream,
    Function,
    Procedure,
    Sequence,
    Table,
]


UTC = "UTC"
APP = "PythonConnector [snowday]"
Q_TAG = "[snowday]"


class Connector:
    def __init__(self, **connection_kwargs):
        self.verbose = True
        for k, v in connection_kwargs.items():
            if k.lower() != "password":
                setattr(self, k, v)
            else:
                setattr(self, k, "*" * len(v))
        self.query_tag = Q_TAG
        self.search_path = None
        self.connection = sf_connector.connect(
            application=APP,
            internal_application_name=APP,
            session_parameters={"QUERY_TAG": Q_TAG},
            **connection_kwargs,
        )
        self.cursor = self.connection.cursor(sf_connector.DictCursor)

        def snowstopper() -> None:  # pragma: no cover
            self.cursor.close()
            self.connection.close()

        atexit.register(snowstopper)

    def __enter__(self, **kwargs):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.connection.commit()

    def enable_verbose(self) -> None:
        self.verbose = True

    def disable_verbose(self) -> None:
        self.verbose = False

    def execute_safe(self, sql: str) -> dict:
        try:
            if self.verbose:
                LOG.info(sql)
            return self.cursor.execute(sql)
        except Exception as e:
            self.cursor.execute("rollback;")
            if self.verbose:
                LOG.error(sql)
                LOG.debug(e)
            raise e

    def use(self, subject: Union[Warehouse, Database, Schema]) -> dict:
        """A way to quickly navigate the primary search path.
        """
        if subject.__class__ == Warehouse:
            self.warehouse = subject.fqn
        else:
            self.search_path = subject.fqn
        sql = f"use {subject.resource_type} {subject.fqn};"
        return self.fetch(sql)

    def set_session_params(self, **session_params) -> None:
        for param, val in session_params.items():
            self.execute_safe(f"alter session set {param} = '{val}';")

    def set_query_tag(self, query_tag) -> None:
        q_tag = f"{query_tag} {Q_TAG}"
        self.query_tag = q_tag
        self.set_session_params(query_tag=q_tag)

    def set_utc(self) -> None:
        self.set_session_params(timezone=UTC)

    def fetchone(self, sql: str) -> dict:
        return self.execute_safe(sql).fetchone()

    def fetch(self, sql: str) -> dict:
        """An alias for fetchone
        """
        return self.fetchone(sql)

    def fetchall(self, sql: str) -> list:
        return self.execute_safe(sql).fetchall()

    def fetchone_by_key(self, sql: str, key: str):
        return self.fetchone(sql)[key]

    def fetch_by_key(self, sql: str, key: str):
        """An alias for fetchone_by_key
        """
        return self.fetchone_by_key(sql, key)

    def fetchall_by_key(self, sql: str, key: str):
        return [row[key] for row in self.fetchall(sql)]

    def yield_all(self, sql: str) -> Iterator[dict]:
        for row in self.fetchall(sql):
            yield row

    def yield_all_by_key(self, sql: str, key: str) -> Iterator:
        for row in self.fetchall(sql):
            yield row[key]

    def commit(self) -> dict:
        return self.cursor.execute("commit;").fetchone()

    def rollback(self) -> dict:
        return self.cursor.execute("rollback;").fetchone()

    def create(self, obj: ALL_OBJS) -> dict:
        return self.fetch(obj.create)

    def create_if_not_exists(self, obj: ALL_OBJS) -> dict:
        return self.fetch(obj.create_if_not_exists)

    def create_or_replace(self, obj: ALL_OBJS) -> dict:
        return self.fetch(obj.create_or_replace)

    def drop(self, obj: ALL_OBJS) -> dict: # FIXME! Add force= param to this?
        # if force: # FIXME! Do this.
        return self.fetch(obj.drop)

    def describe(self, obj: ALL_OBJS) -> dict:
        return self.fetch(obj.describe)



from dataclasses import dataclass
from snowday.acl.grants.util import get_privilege
from snowday.objects.account import Role


@dataclass
class BaseGrant:
    grantee: Role

    @property
    def privilege(self):
        return get_privilege(self.__class__.__name__)

    @property
    def grant(self):
        return f"grant {self.privilege} on {self.subject.resource_type} {self.subject.fqn} to role {self.grantee.name};"

    @property
    def revoke(self):
        return f"revoke {self.privilege} on {self.subject.resource_type} {self.subject.fqn} from role {self.grantee.name};"




from datetime import date, datetime
from dataclasses import dataclass, field
import json
from typing import List, Union
from snowday.objects.schema import SchemaObject
from snowday.exceptions import (
    InvalidWhitespaceException,
    InvalidColumnConstraintException,
)


def get_default_sql(obj) -> str:
    if hasattr(obj, "default"):
        if obj.default == False:  # Has to be first because... false is falsey
            return f" default {obj.default}"
        if obj.default:  # But `None` should be politely excused
            if obj.datatype in ["varchar", "char", "string", "text"]:
                return f" default '{obj.default}'"
            elif obj.datatype == "boolean":
                return f" default {obj.default}"
            elif obj.datatype == "date":
                return f" default cast('{obj.default}' as date)"
            elif obj.datatype in ["datetime", "timestamp", "timestamp_ntz"]:
                return f" default cast('{obj.default}' as timestamp_ntz(9))"
            elif obj.datatype == "time":
                return f" default cast('{obj.default}' as time(9))"
            elif obj.datatype == "timestamp_ltz":
                return f" default cast('{obj.default}' as timestamp_ltz(9))"
            elif obj.datatype == "timestamp_tz":
                return f" default cast('{obj.default}' as timestamp_tz(9))"
            elif obj.datatype == "variant":
                return f" default parse_json('{json.dumps(obj.default)}')"
            elif obj.datatype == "object":
                return f" default parse_json('{json.dumps(obj.default)}')"
            elif obj.datatype == "array":
                return f" default parse_json('{json.dumps(obj.default)}')"
            return f" default {obj.default}"
    return ""


@dataclass
class Column:
    """A base class for abstracting up sql consistencies.
    """

    name: str
    comment: str = None
    # inline constraints
    unique: bool = False
    not_null: bool = False
    # foreign_key: Column  # FIXME! do this!

    def __post_init__(self):
        # Make sure columns don't have whitespace.
        if " " in self.name:
            msg = f"Column names cannot have whitespace. Please change column name {self.name} to {self.name.replace(' ', '_')} or equivalent."
            raise InvalidWhitespaceException(msg)
        # Ensure names are uppercased.
        self.name = self.name.upper()
        # Make sure columns don't have both a unique constraint and a column default.
        if hasattr(self, "default"):
            if self.default and self.unique:
                raise InvalidColumnConstraintException(
                    "A column cannot have both a default and a unique constraint. Please alter one or the other."
                )

    @property
    def datatype(self):
        return "column"

    @property
    def datatype_modifier(self):
        return ""

    @property
    def unique_sql(self):
        if self.unique:
            return " unique"
        return ""

    @property
    def not_null_sql(self):
        if self.not_null:
            return " not null"
        return ""

    @property
    def collation_sql(self):
        # FIXME! Incorporate this functionality
        return ""

    # @property
    # def primary_key_sql(self): # FIXME! do this!
    #     if self.primary_key:
    #         return " primary key "
    #     return ""

    # @property
    # def foreign_key_sql(self): # FIXME! do this!
    #     if self.foreign_key

    @property
    def default_sql(self):
        return get_default_sql(self)

    @property
    def comment_sql(self):
        if self.comment:
            return f" comment '{self.comment}'"
        return ""

    @property
    def sql(self):
        return f"{self.name} {self.datatype}{self.datatype_modifier}{self.collation_sql}{self.default_sql}{self.not_null_sql}{self.unique_sql}{self.comment_sql}"


@dataclass
class BaseNumberColumn(Column):
    default: float = None
    precision: int = 38
    scale: int = 0

    @property
    def datatype(self):
        return "number"

    @property
    def datatype_modifier(self):
        return f"({self.precision}, {self.scale})"


@dataclass
class BaseIntColumn(Column):
    default: int = None

    @property
    def datatype(self):
        return "integer"


@dataclass
class BaseFloatingPointColumn(Column):
    default: float = None

    @property
    def datatype(self):
        return "float"


@dataclass
class BaseTextColumn(Column):
    default: str = None

    @property
    def datatype(self):
        return "text"


@dataclass
class BaseDateColumn(Column):
    default: date = None

    @property
    def datatype(self):
        return "date"


@dataclass
class BaseDatetimeColumn(Column):
    default: datetime = None

    @property
    def datatype(self):
        return "datetime"


# Numeric Columns


@dataclass
class NumberColumn(BaseNumberColumn):
    default: float = None

    @property
    def datatype(self):
        return "number"


@dataclass
class DecimalColumn(BaseNumberColumn):
    default: float = None

    @property
    def datatype(self):
        return "decimal"


@dataclass
class NumericColumn(BaseNumberColumn):
    default: float = None

    @property
    def datatype(self):
        return "numeric"


# Fixed precision/scale


@dataclass
class IntColumn(BaseIntColumn):
    @property
    def datatype(self):
        return "int"


@dataclass
class BigIntColumn(BaseIntColumn):
    @property
    def datatype(self):
        return "bigint"


@dataclass
class SmallIntColumn(BaseIntColumn):
    @property
    def datatype(self):
        return "smallint"


@dataclass
class TinyIntColumn(BaseIntColumn):
    @property
    def datatype(self):
        return "tinyint"


@dataclass
class ByteIntColumn(BaseIntColumn):
    @property
    def datatype(self):
        return "byteint"


@dataclass
class FloatColumn(BaseFloatingPointColumn):
    @property
    def datatype(self):
        return "float"


@dataclass
class Float4Column(BaseFloatingPointColumn):
    @property
    def datatype(self):
        return "float4"


@dataclass
class Float8Column(BaseFloatingPointColumn):
    @property
    def datatype(self):
        return "float8"


@dataclass
class DoubleColumn(BaseFloatingPointColumn):
    @property
    def datatype(self):
        return "double"


@dataclass
class RealColumn(BaseFloatingPointColumn):
    @property
    def datatype(self):
        return "real"


# Text Columns


@dataclass
class VarcharColumn(BaseTextColumn):
    length: int = 16777216

    @property
    def datatype(self):
        return "varchar"

    @property
    def datatype_modifier(self):
        return f"({self.length})"


@dataclass
class CharColumn(BaseTextColumn):
    @property
    def datatype(self):
        return "char"


@dataclass
class StringColumn(BaseTextColumn):
    length: int = 16777216

    @property
    def datatype(self):
        return "string"

    @property
    def datatype_modifier(self):
        return f"({self.length})"


@dataclass
class TextColumn(BaseTextColumn):
    length: int = 16777216

    @property
    def datatype(self):
        return "text"

    @property
    def datatype_modifier(self):
        return f"({self.length})"


@dataclass
class BinaryColumn(BaseTextColumn):
    @property
    def datatype(self):
        return "binary"


@dataclass
class VarbinaryColumn(BaseTextColumn):
    @property
    def datatype(self):
        return "varbinary"


# Logical Columns
@dataclass
class BooleanColumn(Column):
    default: bool = None

    @property
    def datatype(self):
        return "boolean"


# Date & Time Columns
@dataclass
class DateColumn(BaseDateColumn):
    @property
    def datatype(self):
        return "date"


@dataclass
class DatetimeColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "datetime"


@dataclass
class TimeColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "time"


@dataclass
class TimestampColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "timestamp"


@dataclass
class TimestampLTZColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "timestamp_ltz"


@dataclass
class TimestampNTZColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "timestamp_ntz"


@dataclass
class TimestampTZColumn(BaseDatetimeColumn):
    @property
    def datatype(self):
        return "timestamp_tz"


# Semi-structured Columns
@dataclass
class VariantColumn(Column):
    default: Union[list, dict] = None

    @property
    def datatype(self):
        return "variant"


@dataclass
class ObjectColumn(Column):
    default: dict = None

    @property
    def datatype(self):
        return "object"


@dataclass
class ArrayColumn(Column):
    default: list = None

    @property
    def datatype(self):
        return "array"


# Geospatial Columns
@dataclass
class GeographyColumn(Column):
    default: Union[
        dict, str
    ] = None  # FIXME: string must be in WKT, WKB, EWKT, EWKB format

    @property
    def datatype(self):
        return "geography"


ACCEPTED_TABLE_COLUMNS = Union[
    NumberColumn,
    IntColumn,
    BigIntColumn,
    SmallIntColumn,
    TinyIntColumn,
    ByteIntColumn,
    FloatColumn,
    Float4Column,
    Float8Column,
    DoubleColumn,
    RealColumn,
    VarcharColumn,
    CharColumn,
    StringColumn,
    TextColumn,
    BooleanColumn,
    DateColumn,
    DatetimeColumn,
    TimeColumn,
    TimestampColumn,
    TimestampLTZColumn,
    TimestampNTZColumn,
    TimestampTZColumn,
    VariantColumn,
    ObjectColumn,
    ArrayColumn,
    GeographyColumn,
]


@dataclass
class Table(SchemaObject):
    columns: List[ACCEPTED_TABLE_COLUMNS] = None
    cluster_keys: List[ACCEPTED_TABLE_COLUMNS] = None

    def __post_init__(self):
        for column in self.columns:
            setattr(self, f"_{column.name}", column)

    @property
    def resource_type(self):
        return "table"

    @property
    def _definition_sql(self):
        col_sql = ", ".join(c.sql for c in self.columns)
        return f"({col_sql})"

    @property
    def create(self):
        return f"create {self.resource_type} {self.fqn} {self._definition_sql};"

    @property
    def create_if_not_exists(self):
        return f"create {self.resource_type} if not exists {self.fqn} {self._definition_sql};"

    @property
    def create_or_replace(self):
        return (
            f"create or replace {self.resource_type} {self.fqn} {self._definition_sql};"
        )


@dataclass
class ExternalTable:
    # FIXME! Do this!
    pass



from dataclasses import dataclass
from snowday.objects.account import Role
from snowday.acl.grants.base import BaseGrant
from snowday.acl.grants.util import get_privilege


@dataclass
class AccountGrant(BaseGrant):
    @property
    def privilege(self):
        return get_privilege(self.__class__.__name__)

    @property
    def grant(self):
        return f"grant {self.privilege} on account to role {self.grantee.name};"

    @property
    def revoke(self):
        return f"revoke {self.privilege} on account from role {self.grantee.name};"


@dataclass
class CreateAccountGrant(AccountGrant):
    pass


@dataclass
class CreateRoleGrant(AccountGrant):
    pass


@dataclass
class CreateUserGrant(AccountGrant):
    pass


@dataclass
class CreateWarehouseGrant(AccountGrant):
    pass


@dataclass
class CreateDatabaseGrant(AccountGrant):
    pass


@dataclass
class CreateIntegrationGrant(AccountGrant):
    pass


@dataclass
class CreateSchemaGrant(AccountGrant):
    pass


@dataclass
class CreateShareGrant(AccountGrant):
    pass


@dataclass
class ImportShareGrant(AccountGrant):
    pass


@dataclass
class MonitorUsageGrant(AccountGrant):
    pass


@dataclass
class ManageGrantsGrant(AccountGrant):
    pass



VARCHAR = "varchar"
CHAR = "char"
CHARACTER = "character"
STRING = "string"  # Synonymous with VARCHAR
TEXT = "text"  # Synonymous with VARCHAR

BINARY = "binary"  # No notion of unicode characters, so length always measured in terms of bytes
VARBINARY = "varbinary"  # Synonymous with BINARY



from dataclasses import dataclass
from typing import Union
from snowday.objects.account import Role, User


@dataclass
class RoleGrant:
    # https://docs.snowflake.com/en/sql-reference/sql/grant-role.html
    grantee: Union[Role, User]
    subject: Role

    @property
    def grant(self):
        return f"grant role {self.subject.name} to {self.grantee.resource_type} {self.grantee.name};"

    @property
    def revoke(self):
        return f"revoke role {self.subject.name} from {self.grantee.resource_type} {self.grantee.name};"



DATE = "date"  # No time elements
TIME = "time"  # HH:MI:SS

TIMESTAMP_LTZ = "timestamp_ltz"  # UTC time with specified precision; operations performed in current session's tz
TIMESTAMP_NTZ = "timestamp_ntz"  # Wallclock time with specified precision; operations performed without taking any tz into account
TIMESTAMP_TZ = "timestamp_tz"  # UTC time with associated tz offset; session tz offset of no offset is provided


# Aliases
DATETIME = "datetime"  # Alias of TIMESTAMP_LTZ
TIMESTAMP = "timestamp"  # User-specific alias via TIMESTAMP_TYPE_MAPPING session param; TIMESTAMP_NTZ by default



import re


def get_priv_from_cls(class_name: str, start_pos: int, stop_pos: int):
    """A helper function for turning CamelCase class names
    into snake_case names, splitting said name on the underscore,
    and getting the appropriate grant privilege.
    """
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", class_name)
    camelcase_name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()
    return " ".join(camelcase_name.split("_")[start_pos:stop_pos])


def get_privilege(class_name: str):
    """A helper function for getting the name of a
    class's privilege, based on the assumption that
    the privilege will be included in the name.
    """
    return get_priv_from_cls(class_name, 0, -1)


def get_future_privilege(class_name: str):
    """A helper function for getting the name of a
    class's future privilege, based on the assumption
    that the privilege will be included in the name
    and sandwiched between "future" and "grant".
    """
    return get_priv_from_cls(class_name, 1, -1)



from dataclasses import dataclass, field
from typing import List

from snowday.objects.base import BaseSnowflakeObject
from snowday.objects.database import Schema


@dataclass
class SchemaObject(BaseSnowflakeObject):
    schema: Schema
    comment: str

    @property
    def fqn(self):
        return f"{self.schema.fqn}.{self.name}"


@dataclass
class Integration(SchemaObject):
    pass


@dataclass
class Stage(SchemaObject):
    pass


@dataclass
class Task(SchemaObject):
    pass


@dataclass
class View(SchemaObject):
    pass


@dataclass
class MaterializedView(SchemaObject):
    pass


@dataclass
class MaskingPolicy(SchemaObject):
    pass


@dataclass
class FileFormatTypeOptions:
    pass


@dataclass
class FileFormat(SchemaObject):
    pass


@dataclass
class InternalStage(SchemaObject):
    pass


@dataclass
class ExternalStage(SchemaObject):
    pass


@dataclass
class Pipe(SchemaObject):
    pass


@dataclass
class Stream(SchemaObject):
    pass


@dataclass
class Function(SchemaObject):
    pass


@dataclass
class Procedure(SchemaObject):
    pass


@dataclass
class Sequence(SchemaObject):
    pass



class InvalidParamsException(Exception):
    pass


class InvalidWhitespaceException(Exception):
    pass


class UnsupportedFeatureException(Exception):
    pass


class InvalidColumnConstraintException(Exception):
    pass



# Fixed Point
NUMBER = "number"  # Up to 38 digits, with optional precision and scale
DECIMAL = "decimal"  # Synonymous with NUMBER
NUMERIC = "numeric"  # Synonymous with NUMBER
INT = "int"  # Synonymous with NUMBER, except precision and scale cannot be specified
INTEGER = (
    "integer"  # Synonymous with NUMBER, except precision and scale cannot be specified
)
BIGINT = (
    "bigint"  # Synonymous with NUMBER, except precision and scale cannot be specified
)
SMALLINT = (
    "smallint"  # Synonymous with NUMBER, except precision and scale cannot be specified
)
TINYINT = (
    "tinyint"  # Synonymous with NUMBER, except precision and scale cannot be specified
)
BYTEINT = (
    "byteint"  # Synonymous with NUMBER, except precision and scale cannot be specified
)


# Floating Point
FLOAT = "float"
FLOAT4 = "float4"
FLOAT8 = "float8"
DOUBLE = "double"  # Displayed as FLOAT but stored as DOUBLE
DOUBLE_PRECISION = "double precision"  # Displayed as FLOAT but stored as DOUBLE
REAL = "real"  # Displayed as FLOAT but stored as DOUBLE



import copy
from re import finditer


ATTRS_TO_EXCLUDE = [
    "database",
    "schema",
    "name",
    "or_replace",
    "if_not_exists",
    "transient",
]


def get_resource_type(obj):
    matches = finditer(
        ".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)", obj.__class__.__name__
    )
    case_delineated_words = [m.group(0).lower() for m in matches]
    return " ".join(case_delineated_words)


def settings_str_from(obj, exclude_attrs=ATTRS_TO_EXCLUDE):
    d = copy.deepcopy(obj.__dict__)
    settings = []
    for k, v in d.items():
        if k not in exclude_attrs and k[0] != "_":
            if v:
                if isinstance(v, str):
                    settings.append(f"{k}='{v}'")
                elif isinstance(v, list):
                    vals_str = ", ".join([f"'{val}'" for val in v])
                    settings.append(f"{k}=({vals_str})")
                elif isinstance(v, (bool, int)):
                    settings.append(f"{k}={v}")
                else:
                    settings.append(f"{k}='{v.name.upper()}'")
    return " ".join(settings)



from dataclasses import dataclass
from typing import Union

from snowday.objects.account import Role, User, Warehouse, ResourceMonitor
from snowday.objects.database import Database, Schema
from snowday.objects.schema import (
    InternalStage,
    Stage,
    Function,
    FileFormat,
    Integration,
    Pipe,
    Procedure,
    Sequence,
    Stream,
    Task,
    ExternalStage,
    View,
)
from snowday.objects.table import Table
from snowday.acl.grants.base import BaseGrant


@dataclass
class OwnershipGrant(BaseGrant):
    # https://docs.snowflake.com/en/sql-reference/sql/grant-ownership.html
    subject: Union[
        Role,
        User,
        Warehouse,
        Database,
        Schema,
        Table,
        View,
        Stage,
        FileFormat,
        Stream,
        Task,
        Pipe,
        Integration,
        Function,
        Procedure,
        Sequence,
    ]
    revoke_current_grants = False
    copy_current_grants = True


@dataclass
class AllGrant(BaseGrant):
    subject: Union[Warehouse, Table, Schema, InternalStage, ExternalStage, View]


@dataclass
class UsageGrant(BaseGrant):
    subject: Union[
        Warehouse,
        Database,
        Schema,
        Integration,
        ExternalStage,
        FileFormat,
        Function,
        Sequence,
    ]


@dataclass
class OperateGrant(BaseGrant):
    subject: Union[Warehouse, Task]


@dataclass
class ModifyGrant(BaseGrant):
    subject: Union[Warehouse, Database, Schema, ResourceMonitor]


@dataclass
class MonitorGrant(BaseGrant):
    subject: Union[Warehouse, Database, Schema, ResourceMonitor, Task]


@dataclass
class CreateSchemaGrant(BaseGrant):
    subject: Database


@dataclass
class ImportedPrivilegesGrant(BaseGrant):
    subject: Database



from dataclasses import dataclass
from typing import Union


from snowday.acl.grants.base import BaseGrant
from snowday.acl.grants.util import get_future_privilege
from snowday.objects.account import Role
from snowday.objects.schema import (
    View,
    MaterializedView,
    Stream,
    MaskingPolicy,
    InternalStage,
)
from snowday.objects.table import Table


@dataclass
class SchemaObjectGrant(BaseGrant):
    pass


@dataclass
class FutureSchemaObjectGrant:
    # FIXME! Is this the right way forward? Is there a downside here?
    # What is a better way to incorporate future grants on schema objects?
    #
    grantee: Role

    @property
    def privilege(self):
        return get_future_privilege(self.__class__.__name__)

    @property
    def _plural_subject_type(self, suffix="s"):
        return f"{self.subject.__class__.__name__.lower()}{suffix}"

    @property
    def grant(self):
        return (
            f"grant {self.privilege} "
            f"on future {self._plural_subject_type} "
            f"in {self.namespace.resource_type} {self.namespace.name} "
            f"to {self.subject.resource_type} {self.subject.name};"
        )

    @property
    def revoke(self):
        return (
            f"revoke {self.privilege} "
            f"on future {self._plural_subject_type} "
            f"in {self.namespace.resource_type} {self.namespace.name} "
            f"from {self.subject.resource_type} {self.subject.name};"
        )


@dataclass
class SelectGrant(SchemaObjectGrant):
    subject: Union[Table, View, MaterializedView, Stream]


@dataclass
class FutureSelectGrant(FutureSchemaObjectGrant):
    subject: Union[Table, View, MaterializedView, Stream]
    namespace: Union[Database, Schema]


@dataclass
class InsertGrant(SchemaObjectGrant):
    subject: Table


@dataclass
class FutureInsertGrant(FutureSchemaObjectGrant):
    subject: Table
    namespace: Union[Database, Schema]


@dataclass
class UpdateGrant(SchemaObjectGrant):
    subject: Table


@dataclass
class FutureUpdateGrant(FutureSchemaObjectGrant):
    subject: Table
    namespace: Union[Database, Schema]


@dataclass
class DeleteGrant(SchemaObjectGrant):
    subject: Table


@dataclass
class FutureDeleteGrant(FutureSchemaObjectGrant):
    subject: Table
    namespace: Union[Database, Schema]


@dataclass
class TruncateGrant(SchemaObjectGrant):
    subject: Table


@dataclass
class FutureTruncateGrant(FutureSchemaObjectGrant):
    subject: Table
    namespace: Union[Database, Schema]


@dataclass
class ReferencesGrant(SchemaObjectGrant):
    subject: Table


@dataclass
class FutureReferencesGrant(FutureSchemaObjectGrant):
    subject: Table
    namespace: Union[Database, Schema]


@dataclass
class ReadGrant(SchemaObjectGrant):
    subject: InternalStage


@dataclass
class FutureReadGrant(FutureSchemaObjectGrant):
    subject: InternalStage
    namespace: Union[Database, Schema]


@dataclass
class WriteGrant(SchemaObjectGrant):
    subject: InternalStage


@dataclass
class FutureWriteGrant(FutureSchemaObjectGrant):
    subject: InternalStage
    namespace: Union[Database, Schema]


@dataclass
class ApplyGrant(SchemaObjectGrant):
    subject: MaskingPolicy


@dataclass
class FutureApplyGrant(FutureSchemaObjectGrant):
    subject: MaskingPolicy
    namespace: Union[Database, Schema]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snowday',
 'snowday.acl',
 'snowday.acl.grants',
 'snowday.connector',
 'snowday.graph',
 'snowday.objects',
 'snowday.types',
 'snowday.util']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.14.24,<2.0.0',
 'loguru>=0.5.1,<0.6.0',
 'networkx==2.4',
 'snowflake-connector-python==2.2.9']

setup_kwargs = {
    'name': 'snowday',
    'version': '0.1.3',
    'description': 'Tools for turning Snowflake into a Snow Day. ❄️',
    'long_description': '\nTools for turning [Snowflake](https://www.snowflake.com/) into a Snow Day. ❄️\n\nA little bit of sass with a whole lot of `@dataclass`.\n\n\n![Snowday](snowday.png)\n\n\n### Installation\n\n\n```\npip install snowday\n```\n\n\n### What does it do?\n\n#### Connector\n\n![Snowday Connector 1](readme_gif/conn1.gif)\n\n![Snowday Connector 2](readme_gif/conn2.gif)\n\n\n#### Objects\n\n![Snowday Objects 1](readme_gif/obj1.gif)\n\n![Snowday Objects 2](readme_gif/obj2.gif)\n\n\n#### All together now\n\n![Snowday All Together](readme_gif/alltogether.gif)\n\n\n### License\n\n\nSnowday is licensed under GPLv3.0.\n',
    'author': 'Jacob Thomas',
    'author_email': 'jake@bostata.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bostata/snowday',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.8',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
from typing import List

from snowday.objects.base import BaseSnowflakeObject
from snowday.objects.account import Share
from snowday.util import get_resource_type, settings_str_from


# https://docs.snowflake.com/en/sql-reference/sql/create.html


def get_fqn(db_schema_or_string):
    if isinstance(db_schema_or_string, str):
        return db_schema_or_string
    else:
        return db_schema_or_string.fqn


@dataclass
class Database(BaseSnowflakeObject):
    transient: bool = False
    comment: str = ""
    data_retention_time_in_days: int = 1
    default_ddl_collation: str = None

    @property
    def resource_type(self):
        if self.transient:
            return "transient database"
        return "database"


@dataclass
class DatabaseFromShare(BaseSnowflakeObject):
    provider_account: str
    share: Share

    @property
    def _settings_str(self):
        return f"from share {self.provider_account}.{self.share.name}"

    @property
    def resource_type(self):
        return "database"

    @property
    def create_or_replace(self):
        raise UnsupportedFeatureException(
            "Create or replace for snowflake database from share is unsupported."
        )

    @property
    def create_if_not_exists(self):
        raise UnsupportedFeatureException(
            "Create if not exists for snowflake database from share is unsupported."
        )


@dataclass
class DatabaseReplica(BaseSnowflakeObject):
    src_region_name: str
    src_account_name: str
    src_database_name: str
    auto_refresh_materialized_views: bool = False

    @property
    def _settings_str(self):
        return (
            f"as replica of {self.src_region_name}.{self.src_account_name}.{self.src_database_name} "
            f"auto_refresh_materialized_views_on_secondary={self.auto_refresh_materialized_views}"
        )

    @property
    def resource_type(self):
        return "database"

    @property
    def create_or_replace(self):
        raise UnsupportedFeatureException(
            "Create or replace for snowflake database replica is unsupported."
        )

    @property
    def create_if_not_exists(self):
        raise UnsupportedFeatureException(
            "Create if not exists for snowflake database replica is unsupported."
        )


# See: https://docs.snowflake.com/en/sql-reference/sql/create.html
@dataclass
class Schema(BaseSnowflakeObject):
    database: Database
    transient: bool = False
    with_managed_access: bool = False
    data_retention_time_in_days: int = 1
    default_ddl_collation: str = None
    comment: str = ""

    @property
    def resource_type(self):
        if self.transient:
            return "transient schema"
        return "schema"

    @property
    def fqn(self):
        return f"{self.database.name}.{self.name}"

    def _ddl_collation_string(self):
        if self.default_ddl_collation:
            return f"default_ddl_collation={self.default_ddl_collation} "
        else:
            return ""

    @property
    def _settings_str(self):
        return (
            f"{'with managed access ' if self.with_managed_access else ''}"
            f"data_retention_time_in_days={self.data_retention_time_in_days} "
            f"{self._ddl_collation_string()}"
            f"comment='{self.comment}'"
        )



import os
import sys
from loguru import logger


STDOUT_SINK = {"sink": sys.stdout, "level": "DEBUG"}
FILE_SINK = {
    "sink": "snowday.log",
    "rotation": "1 MB",
    "retention": "1 day",
    "enqueue": True,
    "backtrace": True,
    "compression": "gz",
}

CONFIG = {"handlers": [STDOUT_SINK, FILE_SINK], "extra": {"pkg": "[snowday]"}}


def get_logger(config=CONFIG):
    l = logger
    l.configure(**config)
    return l



from dataclasses import dataclass
from typing import Union

from snowday.acl.grants.base import BaseGrant
from snowday.objects.database import Database, Schema
from snowday.acl.grants.util import get_future_privilege


@dataclass
class SchemaGrant(BaseGrant):
    subject: Schema


@dataclass
class FutureSchemaGrant:
    subject: Database
    grantee: Role

    @property
    def privilege(self):
        return get_future_privilege(self.__class__.__name__)

    @property
    def grant(self):
        return (
            f"grant {self.privilege} on future schemas "
            f"in {self.subject.resource_type} {self.subject.name} "
            f"to role {self.grantee.name};"
        )

    @property
    def revoke(self):
        return (
            f"revoke {self.privilege} on future schemas "
            f"in {self.subject.resource_type} {self.subject.name} "
            f"from role {self.grantee.name};"
        )


@dataclass
class FutureAllSchemaGrant(FutureSchemaGrant):
    @property
    def privilege(self):
        return "all"


@dataclass
class CreateTableGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateTableGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateExternalTableGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateExternalTableGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateViewGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateViewGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateMaterializedViewGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateMaterializedViewGrant(SchemaGrant):
    pass


@dataclass
class CreateMaskingPolicyGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateMaskingPolicyGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateFileFormatGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateFileFormatGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateStageGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateStageGrant(FutureSchemaGrant):
    pass


@dataclass
class CreatePipeGrant(SchemaGrant):
    pass


@dataclass
class FutureCreatePipeGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateStreamGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateStreamGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateTaskGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateTaskGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateSequenceGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateSequenceGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateFunctionGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateFunctionGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateUDFGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateUDFGrant(FutureSchemaGrant):
    pass


@dataclass
class CreateProcedureGrant(SchemaGrant):
    pass


@dataclass
class FutureCreateProcedureGrant(FutureSchemaGrant):
    pass
